




from .rectangle import BaseRectangle



class BaseTrigger(BaseRectangle):


    @property
    def blocking(self):
        return False


    @property
    def stationary(self):
        return True

    @property
    def trigger(self):
        return True

    @property
    def collectible(self):
        return True



    def on_trigger(self,obj):
        pass

    def on_trigger_exit(self,obj):
        pass


    def on_collision(self,obj):

        self.on_trigger(obj)
        self.on_trigger_exit(obj)






from .grid import GridEnvironment


class SpatialEnvironment(GridEnvironment):
    def __init__(self,width = None,height = None,
        objects = None,background_color = (0,0,0),
        callbacks_step = None,
        toroidal = True,
        *args,**kwargs,
        ):

        super().__init__(width = width,height = height,cell_size = 1,show_grid = False,objects = objects,
        background_color = background_color,callbacks_step = callbacks_step,toroidal = toroidal,*args,**kwargs)








import numpy as np
import pygame
import itertools
from PIL import Image

from .base_object import BaseObject
from ..colors import *
from ..assets import make_blob,make_ball,make_arrow



class BaseRectangle(BaseObject):

    def __init__(self,x,y,width = 1,height = 1,
        color = (255,0,0),circle = False,radius = None,
        img_filepath = None,img_transparency = (200, 191, 231),
        img_asset = None,
        img_rotate = False,
        collision_method = "rect",
        ):
        
        super().__init__()

        self._x = x
        self._y = y
        self.width = width
        self.height = height
        self.color = color
        self.circle = circle
        self._radius = radius
        self.img_filepath = img_filepath
        self.img_asset = img_asset
        self.img_rotate = img_rotate

        assert collision_method in ["rect","circle"]
        self._collision_method = collision_method

        if img_transparency is None and img_filepath is not None:
            self.img_transparency = np.array(Image.open(img_filepath))[0,0]
        else:
            self.img_transparency = img_transparency


    def init_internals(self,errors_binding = True):

        if not errors_binding:
            if not self.is_bound():
                return None

        # Create rectangle
        if self.img_filepath is not None:

            # Convert alpha instead of convert with transparent pictures
            self.image = pygame.image.load(self.img_filepath).convert()
            self.image.set_colorkey(self.img_transparency)
            self.rescale_img(width = self.width,height = self.height)

        elif self.img_asset is not None:
            assert self.img_asset in ["blob","ball","arrow"]

            if self.img_asset == "blob":
                img = make_blob(self.color,self.img_transparency)
            elif self.img_asset == "ball":
                lighter_fn = lambda x,t : min([255,int((1+t)*x)]) 
                color1 = self.color
                color2 = tuple([lighter_fn(x,0.6) for x in color1])
                img = make_ball(color1,color2,self.img_transparency)
            elif self.img_asset == "arrow":

                img = make_arrow(self.color,transparency = self.img_transparency)

            # Convert alpha instead of convert with transparent pictures
            self.image = pygame.surfarray.make_surface(img)
            self.image.set_colorkey(self.img_transparency)
            self.rescale_img(width = self.width,height = self.height)

            # Create image copy for rotation
            self.raw_image = self.image.copy()

        else:
            # Sprite init
            self.image = pygame.Surface(self.size)
            self.image.set_colorkey((0,0,0))
            self.image.fill(self.color)



        self.rect = self.get_rect()
        self.mask = pygame.mask.from_surface(self.image)


    def __repr__(self):
        return f"Rectangle({self.x},{self.y},size=({self.width},{self.height}))"

    def get_size_img(self):
        return self.image.get_size()

    def rescale_img(self,width,height = None):
        
        # Safety check to avoid rescaling layers
        if not self.layer:
                
            w,h = self.get_size_img()
            new_w = width * self.cell_size
            new_h = int(h*new_w/w) if height is None else height * self.cell_size
            self.image = pygame.transform.scale(self.image, (new_w,new_h))

    
    def rotate_img(self,angle:float) -> None:
        """Rotate the image by a given angle using Pygame Transform functions
        Documentation available here https://www.pygame.org/docs/ref/transform.html#pygame.transform.rotate
        Used to rotate sprites when turning

        Args:
            angle (float): Input angle in degrees for the rotation
        """

        self.image = pygame.transform.rotate(self.raw_image,angle)

    
    @property
    def collision_method(self):
        assert self._collision_method in ["rect","circle"]
        return self._collision_method


    @property
    def x(self):
        return self._x
    
    @property
    def y(self):
        return self._y




    @x.setter
    def x(self, value):
        self._x = value
        self.rect.x = value * self.cell_size


    @y.setter
    def y(self, value):
        self._y = value
        self.rect.y = value * self.cell_size
    


    @property
    def value(self):
        return 1

    @property
    def name(self):
        return self.__class__.__name__

    def __repr__(self):
        return f"{self.name}({self.x},{self.y})"


    @property
    def pos(self):
        return self.x,self.y


    @property
    def pos_array(self):
        # TODO see if faster not as property but attribute
        xs = range(self.x,self.x + self.width)
        ys = range(self.y,self.y + self.height)

        pos = list(itertools.product(ys,xs))
        return pos



    @property
    def dimensions(self):
        return (
            self.x * self.cell_size,
            self.y * self.cell_size,
            *self.size,
        )

    @property
    def size(self):
        return (
            self.width * self.cell_size,
            self.height * self.cell_size,
        )


    @property
    def radius(self):
        if self._radius is None:
            assert self.width == self.height
            return self.width/2
        else:
            return self._radius


    @property
    def center(self):
        radius = self.radius
        return (
            int(self.x * self.cell_size + radius),
            int(self.y * self.cell_size + radius),
        )


    def get_rect(self,x = None,y = None):
        if x is None: x = self.x
        if y is None: y = self.y
        return pygame.rect.Rect((x*self.cell_size,y*self.cell_size,*self.size))


    # def get_collider(self,x,y):
    #     dimensions = (
    #         x * self.cell_size,
    #         y * self.cell_size,
    #         self.width * self.cell_size,
    #         self.height * self.cell_size
    #     )
    #     return pygame.rect.Rect(dimensions)


    # def get_sprite(self,x,y):
    #     sprite = self.sprite
    #     sprite.rect = self.get_collider(x,y)
    #     return sprite




    @property
    def collider(self):
        return pygame.rect.Rect(self.dimensions)


    # @property
    # def sprite(self):
    #     sprite = pygame.sprite.Sprite()
    #     sprite.image = pygame.Surface(self.dimensions[-2:])
    #     sprite.image.set_colorkey((0,0,0))
    #     sprite.image.fill(self.color)
        
    #     # Prepare mask and rect
    #     sprite.rect = self.collider
    #     sprite.mask = pygame.mask.from_surface(sprite.image)
    #     return sprite

        
        

    @property
    def cell_size(self):
        if hasattr(self,"_cell_size"):
            return self._cell_size
        else:
            raise Exception("Object must be added to an environment first to setup box size")


    def bind(self,env):
        self._env = env
        self._cell_size = env.cell_size
        self.init_internals()





    def get_data(self):
        data = {
            "id":self.id,
            "pos":self.pos,
            "type":self.__class__.__name__
        }
        
        if hasattr(self,"attrs"):
            data = {
                **data,
                **{key:getattr(self,key) for key in self.attrs}
            }
        
        return data



    #=================================================================================
    # MOVEMENT
    #=================================================================================

    def move(self,*args,**kwargs):
        pass

    def step(self,*args,**kwargs):
        pass


    # def collides_with(self,others,x = None,y = None):
    #     """Compute collisions between the object and any other objects
    #     Returns if there is a collision + the list of object ids in collision
    #     """

    #     # In the absence of external colliders
    #     # We take the internal collider
    #     if x is None and y is None:
    #         collider = self.collider
    #         sprite = self.sprite
    #     else:
    #         collider = self.get_collider(x,y)
    #         sprite = self.get_sprite(x,y)

        
    #     # If no other colliders
    #     if len(others) == 0:
    #         collisions = []

    #     # Compute collisions using PyGame
    #     else:
    #         other_colliders = [(other.id,other.collider) for other in others if (other.id != self.id and other.blocking)]
    #         if len(other_colliders) > 0:
    #             ids,other_colliders = list(zip(*other_colliders))
    #             collisions = collider.collidelistall(other_colliders)
    #             collisions = [ids[i] for i in collisions]
    #         else:
    #             collisions = []

    #     # Compute collisions with obstacle layer group using pixel perfect collision
    #     if self.env.has_layers:
    #         layer_collisions = pygame.sprite.spritecollide(sprite,self.env._obstacle_layer_group,False,pygame.sprite.collide_mask)
    #         if len(layer_collisions) > 0:
    #             collisions.append("obstacle_layer")

    #     # Return signal of collision and colliders touched
    #     if len(collisions) > 0:
    #         return True,collisions
    #     else:
    #         return False,collisions





    #=================================================================================
    # RENDERERS
    #=================================================================================


    def prerender(self):
        pass

    def postrender(self):
        pass
    

    def render(self,screen = None):
        """Render function to visualize object in the environment
        Visualize either a circle or square in any color
        Vision range can be visualized as well

        TODO: 
            - Improve visualization not only within PyGame
            - Visualize with Sprites as well in PyGame 
        """

        if screen is None:
            screen = self.env.screen


        if self.img_filepath is None and self.img_asset is None:

            if not self.circle:
                # Draw a rectangle on the grid using pygame
                self.render_rect(screen = screen,color=self.color)
                # pygame.draw.rect(screen,self.color,self.dimensions)

            else:
                self.render_circle(screen = screen,color=self.color)
                # Draw a circle on the grid using pygame
                # pygame.draw.circle(screen,self.color,self.center,self.radius)
    
        else:

            if self.img_rotate:
                self.rotate_img(self.angle)

            self.render_img(screen = screen)


        if hasattr(self,"show_search_radius") and self.show_search_radius:
            if hasattr(self,"search_radius_method") and self.search_radius_method == "rect":
                self.render_rect(screen = screen,color = WHITE,radius = self.search_radius,thickness = 1,center = True)
            else:
                self.render_circle(screen = screen,color = WHITE,radius = self.search_radius,thickness = 1)





import sys
sys.path.append("C:/git/reinforcement-learning")


from hyperion.agents import *
from hyperion.environment import *

import random
import numpy as np
import uuid
import attr


STATUSES = ["EGG","CHICKEN","COW","FARMER","SUPERMAN"]
SIZE = 100


@attr.s(slots = True)
class Player(Agent):

    # # Agent id
    # id = attr.ib()
    # id.default
    # def _init_id(self):
    #     return str(uuid.uuid1())

    # Status
    status = attr.ib(default = 0,init=False)

    # Position
    x = attr.ib(init = False)
    @x.default
    def _init_x(self):
        return random.randint(0,SIZE)


    def step(self,env):

        # Movement
        new_x = self.x + random.choice([-1,1])
        new_x = np.clip(new_x,0,SIZE-1)
        self.x = new_x

        # Others
        others = env.inverse_loc(self.id)
        for other in others:
            if other.x == self.x:
                if other.status == self.status:
                    other.status = 0
                    self.status += 1

    def interacts_with(self,other):
        return self.x == other.x,1


class ChickenGame(Environment):

    def render(self):
        env = [" "]*SIZE
        for agent in self.agents:
            env[agent.x] = str(agent.status)
        return "|"+"".join(env)+"|"



    def interactions(self):
        pass







import numpy as np
import random
import pygame
from collections import defaultdict

from ..objects.rectangle import BaseRectangle
from ..algorithms.pathfinding.astar import AStar
from ..algorithms.neighbors import NeighborsFinder


class BaseAgent(BaseRectangle):
    def __init__(self,x,y,width = 1,height = 1,color = (255,0,0),circle = False,diagonal = False,
    curiosity = 20,search_radius = 2,speed = 5,
    active_pathfinding = 1.0,
    **kwargs,
    ):
        super().__init__(x,y,width,height,color,circle,**kwargs)


        # Movement description
        self.speed = speed
        self.diagonal = diagonal
        self.pathfinder = AStar(active_pathfinding)

        # Other characteristics
        self.set_direction()
        self.curiosity = curiosity
        self.search_radius = search_radius
        self._counter = defaultdict(int)
 

    @property
    def stationary(self):
        return False

    def starts_counting(self,name):
        self._counter[name] = 1

    def counter(self,name):
        return self._counter[name]

    def counter_step(self):
        for key,value in self._counter.items():
            self._counter[key] += 1


    def log(self,d):
        d = {"id":self.id,**d}
        self.env.log(d)

    #=================================================================================
    # MOVEMENTS
    #=================================================================================



    def set_direction(self,angle = None):
        if angle is None: angle = np.random.uniform(0,360)
        self.angle = angle


    def turn(self,angle):
        self.angle += angle


    def explore(self):
        pass


    def when_blocked(self,collisions):
        pass

    def wander(self):

        if self.on_grid:
            # Move using an unit movement (adjacent squares only) but more or less in the direction given by the angle
            dx,dy = self._sample_unit_movement_from_angle(self.angle)
            self.move(dx = dx,dy = dy)

        else:
            self.move(speed = self.speed)

        # Change direction based on curiosity value
        if self.clock > 1 and self.clock % self.curiosity == 0:
            self.set_direction()



    @staticmethod
    def _sample_unit_movement_from_angle(angle):
        # TODO does not allow diagonal movement for now
        
        # Compute cos and sin
        x = np.cos(angle)
        y = np.sin(angle)
        
        # Test conditions
        if x == 0.0:
            return (0,y)
        elif y == 0.0:
            return (x,0)
        
        # Compute signs
        sign_x = x / np.abs(x)
        sign_y = y / np.abs(y)
        
        # Compute absolute values and probabilities
        x,y = np.abs([x,y])


        #-----------------
        # Other option here is as followed
        # First experiments show it's a little slower (0.1 us)
        # yet it's maybe more generalizable
        # py = y / (x+y)
        
        # # Compute dx and dy based on probabilities
        # dy = np.random.binomial(1,py)
        # dx = 1 - dy
        
        # # Add signs
        # dx *= sign_x
        # dy *= sign_y

        # Draw values between 0 and x,y
        xd = np.random.uniform(0,x)
        yd = np.random.uniform(0,y)
        
        if xd > yd:
            return (sign_x,0)
        else:
            return (0,sign_y)


    def follow_direction(self,dr = 1):
        # Will be more suitable for a navigation agent implemented in a future class
        # Grid agent can only move in adjacent squares, eventually diagonals
        return self.move(angle = self.angle,dr = dr)


    def move_at(self,x,y):
        dx = x - self.x
        dy = y - self.y 
        self.move(dx = dx,dy = dy)



    def find_path_towards(self,x = None,y = None,obj = None,n = None):

        # Start position
        start_pos = self.pos_array[0]

        # Prepare target position
        if obj is not None:
            if isinstance(obj,str):
                obj = self.env[obj]
            target_pos = obj.pos_array[0]
        else:
            target_pos = y,x

        if self.pathfinder.needs_recompute(start_pos,target_pos):

            # Prepare mesh
            mesh = self.env.get_navigation_mesh(self)

            # Find path
            path = self.pathfinder.run(mesh,start_pos,target_pos,diagonal=self.diagonal,n=n)
            return path
        
        else:
            return self.pathfinder.last_path


    def move_towards(self,x = None,y = None,obj = None,n = None,naive = False):
        """Movement function, during one step the agent will move towards a target position or object using pathfinding
        """


        if obj is None:
            if self.pos == (x,y):
                return True
        else:

            # If given object is a string, we retrieve the object from the inventory using the id
            # Returns an error if the object id is not in the inve  ntory
            if isinstance(obj,str): obj = self.env[obj]

            if self.pos == obj.pos:
                return True

            x,y = obj.pos


        # TODO add in naive pathfinding
        if naive:
            MOVES = []
            if self.x > x:
                MOVES.append((-1,0))
            elif self.x < x:
                MOVES.append((1,0))
            if self.y > y:
                MOVES.append((0,-1))
            elif self.y < y:
                MOVES.append((0,1))

            dx,dy = random.choice(MOVES)
            self.move(dx,dy)

        else:

            # Find path with pathfinding algorithm
            path = self.find_path_towards(x,y,obj,n)

            # Move to next position in the path
            # First element is the current position
            # Warning coordinates are reversed in numpy arrays
            if path is not None:
                if len(path) > 1:
                    y,x = path[1]
                    self.move_at(x,y)
                    return False
                else:
                    return False


    def follow_mouse(self,n = None):
        """Movement function, during one step the agent will follow the mouse position using pathfinding
        """
        x,y = pygame.mouse.get_pos()
        x = x // self.cell_size
        y = y // self.cell_size
        self.move_towards(x = x,y = y,n = n)


    def move(self,dx:int = 0,dy:int = 0,speed:float = None,angle:float = None) -> bool:
        """Movement main function, moves an agent to a new location
        Movement can either be launched with:
            1/ a delta x and y (useful in grid environments) (if speed is None)
            2/ Or using a given speed and angle (useful in spatial environments) (if speed is given)

        The movement will be blocked if the object has a blocking property == True
        and enters a collision given by self.collides()

        Args:
            dx (int, optional): Delta x for the movement. Defaults to 0.
            dy (int, optional): Delta y for the movement. Defaults to 0.
            speed (float, optional): Speed for the angular movement. Defaults to None.
            angle (float, optional): Angle for the movement. Defaults to None.

        Returns:
            bool: [description]
        """

        # Store default values for collisions
        is_collision = False

        if speed is not None:
            if angle is None: angle = self.angle
            angle = 2* np.pi * angle / 360 #- 0.5 * np.pi
            dx = int(speed * np.cos(angle))
            dy = - int(speed * np.sin(angle)) # Negative because y axis is inverted

            self.move(dx = dx,dy = dy)

        # Move using radial movement (with angle and radius)
        elif angle is not None:

            cell_size = self.cell_size

            # Compute delta directions with basic trigonometry
            # In a grid environment, movement is rounded to the integer to fit in the grid
            dx = int(speed * cell_size * np.cos(angle))
            dy = int(speed * cell_size * np.sin(angle))

            return self.move(dx = dx,dy = dy)

        # Move using euclidean movement (with dx and dy)
        else:

            # Store old position
            old_pos = self.x,self.y
                
            # Store movements
            x = int(self.x + dx)
            y = int(self.y + dy)

            # Correct moves going offscreen
            # TODO use this function only in toroidal environments
            x,y = self.env.wrap_in_toroidal_envs(x,y)

            # Store new positions as attributes
            self.x = x
            self.y = y

            # Compute collisions
            collisions = self.collides()
            is_collision = len(collisions) > 0

            # If in collision, cancel movement and trigger block callback
            if is_collision:
                self.x,self.y = old_pos
                self.when_blocked(collisions)

        return is_collision




    def random_walk(self):

        if self.diagonal:

            # Sample a random move between -1, 0 or + 1
            # ie 8 moves around agent's position
            dx,dy = np.random.randint(0,3,2) - 1

        else:
            dx,dy = random.choice([(1,0),(-1,0),(0,1),(0,-1)])


        self.move(dx,dy)



    #=================================================================================
    # FINDER
    #=================================================================================


    def find_closest(self,k = 1,**kwargs):

        # Get objects data we want to search
        objects_data = self.env.find(return_data = True,**kwargs)
        objects_data = objects_data.drop(self.id,errors = "ignore")
        
        if len(objects_data) == 0:
            distances,ids = [],[]
        else:
            # Create neighbors algorithm
            finder = NeighborsFinder(objects_data)
            distances,ids = finder.find_closest(self,k = k)

        return ids


    def find(self,return_objects = True,**kwargs):
        return self.env.find(return_objects = return_objects,**kwargs)


    def find_in_range(self,radius = None,method = "circle",**kwargs):
        # TODO add parameters to only find objects that matters
        # For example not using obstacles
        # Easily done by appending to condition to add obstacles = False

        if radius is None:
            radius = self.search_radius
            assert radius is not None

        group = pygame.sprite.Group()
        
        objs = self.env.find(return_objects = True,**kwargs)
        group.add(*objs)

        search = self.collides_group(group,method = method,ratio = radius)
        return search


        # # Find objects data
        # objects_data = self.env.find_objects(condition = condition,return_data = True)
        # objects_data = objects_data.drop(self.id,errors = "ignore")

        # if len(objects_data) == 0:
        #     return []
        # else:
        #     # Create finder algorithm
        #     finder = NeighborsFinder(objects_data)
        #     ids = finder.find_in_range(self,search_range)

        #     return ids



import numpy as np

def make_blob(color = (0, 162, 232),transparency = (200, 191, 231)):

    return np.array(
        [[transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [255, 255, 255],
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        [63, 72, 204],
        [63, 72, 204],
        [63, 72, 204],
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        [255, 255, 255],
        [255, 255, 255],
        [255, 255, 255],
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        color,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency]]
    ).swapaxes(0,1)


import numpy as np
def make_arrow(color = (63,72,204),color2 = None,transparency = (200,191,231)):

    if color2 is None:
        color2 = color

    return np.flip(np.array(
        [[transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         color2,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency],
         [transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency,
         transparency]]
),axis = 0)




"""Layer Class

Three types of layers
- Obstacle Layer
- Trigger Layer
- Render Layer (does nothing more than display content)
"""

# Maybe inherit from Sprite also
# To be able to add in groups


import pygame
import time
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
from ipywidgets import interact,FloatSlider,IntSlider


from .rectangle import BaseRectangle
from ..utils.image import snap_mask_to_grid,image3d_to_mask,mask_to_image3d,mask_to_mesh
from ..exceptions import ObjectNotBoundError


class BaseLayer(BaseRectangle):
    def __init__(self,img_filepath,img_transparency = (255,255,255),obstacle = True,trigger = False,mask_threshold = 0.1,init_window = True,**kwargs):

        super().__init__(x = 0,y = 0,img_filepath = img_filepath,img_transparency = img_transparency,**kwargs)

        self._obstacle = obstacle
        self._trigger = trigger
        self._raw_img = Image.open(self.img_filepath)
        self.mask_threshold = mask_threshold
        self.saved_cell_size = 0
    
    def init_internals(self):
        super().init_internals()
        self.mask_array = self.get_mask()
        self.raw_img = np.copy(self.get_img())

    def get_size(self):
        return self._raw_img.size


    @property
    def layer(self):
        return True


    @property
    def obstacle(self):
        return self._obstacle
    
    @property
    def trigger(self):
        return self._trigger


    def save_img(self):
        """Save transformed layer as png for further use
        """
        name = f"Layer_{str(time.time())[:10]}_cellsize={self.saved_cell_size}.png"
        Image.fromarray(self.get_img()).save(name)
        print(f"Saved image as {name}")


    def get_img(self):
        """Get numpy array image from pygame surface
        Axes are swapped between numpy and pygame
        """
        if not hasattr(self,"image"):
            raise ObjectNotBoundError("You need to bind the layer to an environment first, because self.image is not initialized")
        else:
            return pygame.surfarray.array3d(self.image).swapaxes(0,1)

    def get_mask(self):
        """Get mask from image, ie a 2D array with 0 for transparent and 1 for object
        """
        img = self.get_img()
        return image3d_to_mask(img)


    def snap_to_grid(self,threshold = None,cell_size = None,return_3d = False):
        """Snap a mask to a given grid, can be used to transform a layer for a grid environment

        Args:
            threshold (float, optional): Threshold to consider a mask, 0.1 means a box with 10% of 1 will be a 1 . Defaults to None, which will use the attribute .mask_threshold
            cell_size (int, optional): the size for each cell in the grid. Defaults to None, which will take the environment value env.cell_size
            return_3d (bool, optional): Returns the array as 3D numpy array or 2D mask. Defaults to False.

        Returns:
            np.ndarray: The mask snapped to a grid
        """

        if threshold is None: threshold = self.mask_threshold
        if cell_size is None: cell_size = self.env.cell_size

        mask = snap_mask_to_grid(self.mask_array,cell_size,threshold)
        if return_3d:
            return mask_to_image3d(mask)
        else:
            return mask


    def set_sprite_image(self,img):
        """Set sprites image from a numpy array
        Useful when raw image is transformed using thresholds
        """

        surface = pygame.surfarray.make_surface(img.swapaxes(0,1))
        self.image = surface


    def show(self):
        """Show the current layer in the notebook using matplotlib
        """
        plt.imshow(self.get_img())
        plt.xticks([])
        plt.yticks([])
        plt.show()



    def explore_snap_to_grid(self,save = True):
        """Explore mask transformation snapped to a grid using ipywidgets
        Can be used in the notebook
        Save argument will actually transforms inplace the layer to be used, 
        But we recommend to save and later use .save_img to save a finalized png version of the layer

        Args:
            save (bool, optional): Save cell_size, image, and threshold to attributess. Defaults to True.
        """


        @interact(
            th = FloatSlider(min = 0,max = 1,value = 0.1,step = 0.1),
            cell_size = IntSlider(min = 10,max = 200,value = 20,step = 10)
        )
        def show(th,cell_size):

            # Prepare figure
            fig, (ax1,ax2) = plt.subplots(1, 2,figsize = (8,6))

            # Snap image mask to grid
            img = self.snap_to_grid(threshold = th,cell_size = cell_size,return_3d = True)

            # Show using matplotlib
            ax1.imshow(self.raw_img)
            ax2.imshow(self.snap_to_grid(threshold = th,cell_size = cell_size,return_3d = True))

            # Add titles and removing ticks
            ax1.set_title("Original image")
            ax2.set_title(f"Snap to grid\n'threshold={th}' & 'cell_size={cell_size}'")
            ax1.set_xticks([])
            ax2.set_xticks([])
            ax1.set_yticks([])
            ax2.set_yticks([])
            plt.show()

            # Save last params
            if save:
                self.mask_threshold = th
                self.saved_cell_size = cell_size
                self.set_sprite_image(img)


    def get_navigation_mesh(self):
        """Get a navigation mesh snapped to the grid
        """
        shape = (self.env.height,self.env.width)
        if self.obstacle:
            mesh = mask_to_mesh(self.mask_array,self.env.cell_size,threshold=self.mask_threshold)
            assert mesh.shape == shape
            return mesh
        else:
            return np.zeros(shape)





import os
import numpy as np
import matplotlib.pyplot as plt
from mazelib import Maze
from mazelib.generate.DungeonRooms import DungeonRooms
from mazelib.generate.Prims import Prims

from ipywidgets import interact,Dropdown,IntSlider
from PIL import Image


from .image import mesh_to_mask,mask_to_image3d



class MazeGenerator:
    """Helper class to generate mazes using mazelib
    """

    def __init__(self,width = 20,height = 20,cell_size = 20):
        
        self.width = width
        self.height = height
        self.cell_size = cell_size
        self.maze = Maze()
        self.options = ["dungeon","prims"]

    @property
    def mesh(self):
        return self.maze.grid


    def make_mask(self):
        mask = mesh_to_mask(self.mesh,self.cell_size)
        # return mask
        img = mask_to_image3d(mask)#.astype(np.int8)
        img = np.asarray(img,dtype=np.uint8)
        return img


    def generate(self,method = ["dungeon"],rooms = None,width = None,height = None):

        if width is None: width = self.width
        if height is None: width = self.height

        width = width // 2
        height = height // 2

        if method == "dungeon":
            self.maze.generator = DungeonRooms(width,height, rooms=rooms)
        elif method == "prims":
            self.maze.generator = Prims(width,height)
        else:
            raise Exception(f"Maze generation method must be in {self.options}")

        self.maze.generate()



    def explore(self,**kwargs):


        @interact(
            method = Dropdown(desc = "Generation method",options = self.options),
            i = IntSlider(desc = "Generation",min = 0,max = 100,value = 0),
            width = IntSlider(desc = "Width",min = 10,max = 300,value = 20),
            height = IntSlider(desc = "Height",min = 10,max = 300,value = 20),
        )
        def show(method,i,width,height):

            self.generate(method = method,width = width,height = height,**kwargs)
            mesh = self.mesh
            plt.figure(figsize = (8,8))
            plt.imshow(mesh)
            plt.title("Generated maze")
            plt.show()



    def save(self,folder = "."):

        filename = f"GeneratedMaze_cellsize={self.cell_size}.png"
        filepath = os.path.join(folder,filename)

        mask = self.make_mask()

        Image.fromarray(mask).save(filepath)
        print(f"... Saved maze in {filepath}")





# Base libraries
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import uuid
import imageio
from tqdm import tqdm_notebook
from collections import defaultdict

# Interaction & animation
from ipywidgets import widgets
from IPython.display import display
from ipywidgets import interact,IntSlider,Text


def action(duration,delay = 0,periods = 1,loop = False):
    t = 0

    while True:
        if t%(duration+delay) < delay:
            start,end = False,False
            yield start,end
        else:
            start,end = True,False
            yield start,end
        t += 1

        if not loop:
            if t//(duration+delay) >= periods:
                break

    start,end = True,True
    yield start,end



#======================================================================================
# ENVIRONMENT CLASSES
#======================================================================================


class Environment:
    def __init__(self):

        SCHEMA = ["agent_id","agent_type","agent"]
        self._data = pd.DataFrame(columns = SCHEMA).set_index("agent_id")


    @property
    def agents(self):
        return self._data["agent"].tolist()

    @property
    def data(self):
        return self._data.drop(columns = "agent")
    
    def __getitem__(self,key):
        if isinstance(key,int):
            return self._data.iloc[key].loc["agent"]
        else:
            return self.get_agent(key)

    def __iter__(self):
        return iter(self.agents)


    def _repr_html_(self):
        return self.data.head(20)._repr_html_()

    def add_agent(self,agent,agent_data):
        agent_data["agent"] = agent
        agent_data = pd.DataFrame([agent_data]).set_index("agent_id")
        self._data = self._data.append(agent_data,verify_integrity = True,sort = False)

    def remove_agent(self,agent_id):
        self._data.drop(agent_id,inplace = True)

    def get_agent(self,agent_id):
        return self._data.loc[agent_id,"agent"]

    def step(self,fast = False):
        """Discrete step function 
        """

        if not fast:

            # Initialize reward at 0
            reward = 0

            # Loop over each agent
            for agent in self.agents:
                reward_agent = agent.step()
                reward += reward_agent

            # Compute if episode is finished
            done = len(self.agents) == 0

            return reward,done

        else:
            reward = self._data["agent"].map(lambda x : x.step()).sum()


    def run(self,n,render = True):
        """Run episode function
        """

        experiment = Experiment()

        # Loop each step in the episode
        for i in tqdm_notebook(range(n)):

            # Compute reward and if episode is finished
            reward,done = self.step()

            # If episode is finished stop
            if done:
                break

            # Otherwise append to reward and save image
            else:

                data = {"reward":reward}

                if render:
                    fig = self.render(return_fig = True)
                    data["fig"] = fig

                experiment.log(data)


        return experiment


def fig_to_img(fig):
    fig.canvas.draw_idle()
    image = np.frombuffer(fig.canvas.tostring_rgb(), dtype='uint8')
    image  = image.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    plt.close()
    return image



class Zone2D:

    def __init__(self,xmin,xmax,ymin,ymax,forbidden = False):

        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.forbidden = forbidden

    def contains(self,x,y):
        return self.contains_x(x) and self.contains_y(y)

    def contains_x(self,x):
        return (self.xmin < x < self.xmax)

    def contains_y(self,y):
        return (self.ymin < y < self.ymax)

    def allow(self,x = None,y = None):

        if x is None:
            contains = self.contains_y(y)
        elif y is None:
            contains = self.contains_x(x)
        else:
            contains = self.contains(x,y)

        if not self.forbidden:
            return contains
        else:
            return not contains









class Environment2D(Environment):
    def __init__(self,zones = None):
        """2D Environment
        bounds = (xmin,xmax,ymin,ymax)
        Add something like an occlusion zone, or at least somewhere where agents can't go
        """
        super().__init__()
        self.zones = zones





    def render(self,return_img = False,return_fig = False):

        # Create figure
        fig = plt.figure(figsize=(7,7))
        ax = fig.add_subplot(111)

        # Set bounds if needed
        if self.zones is not None:
            ax.set_xlim([self.zones[0].xmin,self.zones[0].xmax])
            ax.set_ylim([self.zones[0].ymin,self.zones[0].ymax])

        # Plot scatter plot
        self.data[["x","y"]].plot(kind = "scatter",x="x",y="y",ax = ax)

        # Return image for animation
        if return_img:
            # From https://ndres.me/post/matplotlib-animated-gifs-easily/
            img = fig_to_img(fig)
            return img

        # Return figure
        if return_fig:
            plt.close()
            return fig


        

#======================================================================================
# AGENT CLASSES
#======================================================================================




class Agent:
    def __init__(self,env,agent_data = {}):

        # Base parameters
        self.agent_id = str(uuid.uuid1())
        self.agent_type = self.__class__.__name__
        self.internal_clock = 0

        # Base agent data
        agent_data = {
            "agent_id":self.agent_id,
            "agent_type":self.agent_type,
            **agent_data
        }

        # Prepare data as argument        
        self.env = env
        self.env.add_agent(self,agent_data)


    def step(self):
        self.internal_clock += 1


    def __getitem__(self,key):
        return self.get(key)

    def __repr__(self):
        return f"{self.agent_type}(id={self.agent_id})"

    def get_data(self):
        return self.env[self.agent_id].to_dict()

    def get(self,key):
        return self.env._data.loc[self.agent_id,key]

    def set(self,key,value):
        self.env._data.loc[self.agent_id,key] = value

    def add(self,key,value):
        self.env._data.loc[self.agent_id,key] += value

    def multiply(self,key,value):
        self.env._data.loc[self.agent_id,key] *= value

    def sub(self,key,value):
        self.env._data.loc[self.agent_id,key] -= value

    def divide(self,key,value):
        self.env._data.loc[self.agent_id,key] /= value



class Agent2D(Agent):


    def move(self,x = None,y = None,dx = 0,dy = 0,angle = None):
        """Move function for base agent

        TODO:
        - find a way to do allowed moves programmatically and dynamically
        - Subclass this function into a Agent2D
        """

        # Move if x and y are given
        if x is not None and y is not None:
            pass

        # Move if angle and velocity are given
        elif angle is not None:

            # Compute delta directions with basic trigonometry
            dx = self.velocity * np.cos(angle)
            dy = self.velocity * np.sin(angle)

            # Move towards other point
            self.move(dx = dx,dy = dy)

        # Move with delta directions
        else:

            # Compute new directions
            x = self["x"] + dx
            y = self["y"] + dy

            # Change if allowed
            if self.env.zones is None or all(map(lambda zone : zone.allow(x = x),self.env.zones)):
                self.set("x",x)

            if self.env.zones is None or all(map(lambda zone : zone.allow(y = y),self.env.zones)):
                self.set("y",y)



    def move_towards(self,x_target,y_target):

        # Find coords
        x,y = self.get("x"),self.get("y")

        # Compute direction with basic trigonometry
        angle = np.arctan2(y_target - y,x_target - x)

        # Move towards target
        self.move(angle = angle)


    @staticmethod
    def generator_wander(pivot_frequency):
        t = 0
        while True:
            if t % pivot_frequency == 0:
                angle = np.random.uniform(0,2*np.pi) 
            yield angle
            t += 1

    def wander(self,gen = None,pivot_frequency = 3):

        if not hasattr(self,"_generator_wander"):
            self._generator_wander = self.generator_wander(pivot_frequency)
        if gen is None:
            gen = self._generator_wander

        angle = next(gen)
        self.move(angle = angle)




class StaticAgent(Agent):
    pass







#======================================================================================
# SUGARSCAPE CLASSES
#======================================================================================



class Rabbit(Agent2D):
    def __init__(self,env,**kwargs):

        # Prepare agent parameters
        agent_data = {
            "life":np.random.randint(100,200),
            "x":np.random.randint(0,10),
            "y":np.random.randint(0,10),
            "velocity":0.5,
        }
        agent_data["life_left"] = agent_data["life"]

        # Init
        super().__init__(env,agent_data)

        # Init moves generators
        self._generator_wander = self.generator_wander(pivot_frequency = 5)


    @property
    def velocity(self):
        return self["velocity"]
    

    def step(self):
        super().step()

        self.sub("life_left",1)

        # try:
        #     start,end = next(self.actions)
        #     if end:
        #         print(f"Recharge {self.agent_id}!")
        #         self.add("life_left",10)
        # except:
        #     pass

        # Aleatory move
        # dx,dy = np.random.randn() / 5,np.random.randn() / 5
        # self.move(dx = dx,dy = dy)


        self.wander()

        # self.move_towards(10,10)

        if self["life_left"] == 0:
            self.env.remove_agent(self.agent_id)
            return -1
        else:
            return 0



class Food(StaticAgent):
    def __init__(self):
        pass




class Sugarscape(Environment2D):
    pass



class Experiment:

    def __init__(self):

        self.data = defaultdict(list)


    @property
    def fig(self):
        return self.data["fig"]
    

    def log(self,data):
        for k,v in data.items():
            self.data[k].append(v)

    def save_as_gif(self,path):
        """Save experiment as gif
        """
        pass

        #         if save is not None:
        #             assert isinstance(save,str)
        #             img = self.show(return_img = True)
        #             imgs.append(img)

        # if save is not None:
        #     imageio.mimsave(save,imgs)

    def replay(self):
        """Replay function using ipywidgets and interact
        TODO:
        - Add slider for finetuning
        """

        play = widgets.Play(
            value=0,
            min=0,
            max=len(self.fig)-1,
            step=1,
            description="Press play",
            disabled=False
        )


        @interact(
            i = play,
        )
        def show(i):
            return self.fig[i]





class PathfindingError(Exception):
    pass



class Pathfinding:


    @staticmethod
    def get_adjacent_moves(diagonal = False):

        moves = [(0, -1), (0, 1), (-1, 0), (1, 0)]
        diagonal_moves = [(-1, -1), (-1, 1), (1, -1), (1, 1)]

        if not diagonal:
            return moves
        else:
            return moves + diagonal_moves


    @staticmethod
    def show_path(mesh,start,end,path,sep = " "):

        path_str = []

        for i,row in enumerate(mesh):
            row_str = []
            for j,cell in enumerate(row):
                if (i,j) == start:
                    row_str.append("O")
                elif (i,j) == end:
                    row_str.append("X")
                elif cell == 1:
                    row_str.append("#")
                elif (i,j) in path:
                    row_str.append(".")
                else:
                    row_str.append(" ")

            row_str = ["|"] + row_str + ["|"]
            row_str = sep.join(row_str)
            path_str.append(row_str)

        print("\n".join(path_str))




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['westworld',
 'westworld._deprecated',
 'westworld.agents',
 'westworld.algorithms',
 'westworld.algorithms.pathfinding',
 'westworld.assets',
 'westworld.assets.sprites',
 'westworld.environment',
 'westworld.objects',
 'westworld.optimization',
 'westworld.simulation',
 'westworld.utils']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.4.2,<4.0.0',
 'numpy>=1.21.0,<2.0.0',
 'pandas>=1.3.0,<2.0.0',
 'tqdm>=4.61.2,<5.0.0']

setup_kwargs = {
    'name': 'westworld',
    'version': '0.1.0',
    'description': 'Multi-agent simulation library in Python',
    'long_description': '# Westworld\n![](docs/docs/img/cover_hq_westworld1.jpg)\n\n# Description\n**Westworld** is a multi-agent simulation library, its goal to simulate and optimize systems and environments with multiple agents interacting. Its inspiration is drawn from Unity software and [Unity ML Agents](https://github.com/Unity-Technologies/ml-agents), adapted in Python. \n\nThe goal is to be able to simulate environments in logistics, retails, epidemiology, providing pre-coded spatial environments and communication between agents. Optimization can be included using heuristics as well as Reinforcement Learning.\n\n*Library under active development*\n\n*The name is of course inspired by the TV series Westworld, which is actually a gigantic multi-agent simulation system.*\n\n# Documentation\nDocumentation is available locally in docs folder or online at https://theolvs.github.io/westworld\n\n\n# Repo structure\n\n```\n- docs : documentation of the library\n- examples : simulation examples and development tests\n- westworld : library\n```',
    'author': 'Theo Alves Da Costa',
    'author_email': 'theo.alves.da.costa@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""A* pathfinding algorithm implementation
Closely adapted from the work of : 
- Nicholas Swift https://medium.com/@nicholas.w.swift/easy-a-star-pathfinding-7e6689c7f7b2
- Ryan Collinwood https://gist.github.com/ryancollingwood/32446307e976a11a1185a5394d6657bc (improvement on the latter example)
"""

import numpy as np
import random
import heapq
import warnings

from .pathfinding import Pathfinding,PathfindingError



class Node:
    """A node class for A* Pathfinding"""

    def __init__(self, parent=None, position=None):
        self.parent = parent
        self.position = position

        self.g = 0
        self.h = 0
        self.f = 0

    def __eq__(self, other):
        return self.position == other.position


    def __repr__(self):
        return f"{self.position} - g: {self.g} h: {self.h} f: {self.f}"

    # defining less than for purposes of heap queue
    def __lt__(self, other):
        return self.f < other.f
    
    # defining greater than for purposes of heap queue
    def __gt__(self, other):
        return self.f > other.f



class AStar(Pathfinding):

    def __init__(self,proba_recompute = 1.0):


        # Probability to recompute the path at each step
        # 1 represents active pathfinding where we recompute at each step
        self.proba_recompute = float(proba_recompute)
        assert 0 <= self.proba_recompute <= 1

        self.last_target = (-1,-1)
        self.last_path = []


    def needs_recompute(self,start_pos,target_pos):

        # If active pathfinding
        if self.proba_recompute == 1:
            return True

        # If target is the same, draw a random float to see if we recompute 
        elif target_pos == self.last_target:
            return random.random() < self.proba_recompute

        # If target is different we of course needs to recompute
        else:
            return True


    def return_path(self,current_node):
        path = []
        current = current_node
        while current is not None:
            path.append(current.position)
            current = current.parent

        # Return reversed path
        path = path[::-1]

        # Store last path found
        self.last_path = path
        
        return path


    def run(self,mesh,start,end,diagonal = False,n = None):
        """Returns a list of tuples as a path from the given start to the given end in the given mesh
        Base implementation is drawn from Nicholas Swift work explained at https://medium.com/@nicholas.w.swift/easy-a-star-pathfinding-7e6689c7f7b2

        TODO - This can probably be a lot accelerated by numba and or cython
        TODO - Requires for westworld env to have a get_nav_mesh() method extracting numpy array
        TODO - Will only work if grid environment, yet possible to discretize env if needed
        TODO - Add stop at n iterations to fasten computation
        """

        # Safety check
        try:
            mesh[start],mesh[end]
        except IndexError as e:
            raise PathfindingError(f"Start or end not in the mesh: {e}")

        # Create start and end node
        start_node = Node(None, start)
        end_node = Node(None, end)

        # Store last target
        self.last_target = end

        # Get adjacent moves
        moves = self.get_adjacent_moves(diagonal)

        # Initialize both open and closed list
        open_list = []
        closed_list = []

        # Heapify the open_list and Add the start node
        heapq.heapify(open_list) 
        heapq.heappush(open_list, start_node)

        # Adding a stop condition
        outer_iterations = 0
        max_iterations = np.product(mesh.shape)
        if n is not None:
            max_iterations = min([max_iterations,n])

        # Loop until you find the end
        while len(open_list) > 0:

            outer_iterations += 1

            # Get the current node
            current_node = heapq.heappop(open_list)
            closed_list.append(current_node)

            if outer_iterations > max_iterations:
                # if we hit this point return the path such as it is
                # it will not contain the destination
                warnings.warn("Giving up on pathfinding too many iterations")
                return self.return_path(current_node)     

            # Found the goal
            if current_node == end_node:
                return self.return_path(current_node)

            # Generate children
            children = []
            
            for new_position in moves: # Adjacent squares

                # Get node position
                node_position = (current_node.position[0] + new_position[0], current_node.position[1] + new_position[1])

                # Make sure within range
                if node_position[0] > (len(mesh) - 1) or node_position[0] < 0 or node_position[1] > (len(mesh[len(mesh)-1]) -1) or node_position[1] < 0:
                    continue

                # Make sure walkable terrain
                if mesh[node_position[0]][node_position[1]] != 0:
                    continue

                # Create new node
                new_node = Node(current_node, node_position)

                # Append
                children.append(new_node)

            # Loop through children
            for child in children:
                # Child is on the closed list
                if len([closed_child for closed_child in closed_list if closed_child == child]) > 0:
                    continue

                # Create the f, g, and h values
                child.g = current_node.g + 1
                child.h = ((child.position[0] - end_node.position[0]) ** 2) + ((child.position[1] - end_node.position[1]) ** 2)
                child.f = child.g + child.h

                # Child is already in the open list
                if len([open_node for open_node in open_list if child.position == open_node.position and child.g > open_node.g]) > 0:
                    continue

                # Add the child to the open list
                heapq.heappush(open_list, child)

        warnings.warn("Couldn't get a path to destination")
        return None






from .trigger import BaseTrigger


class BaseCollectible(BaseTrigger):
    """Collectible objects in an environment
    Collectibles are objects an agent can find, after having been found the object disappears
    Eg: Pacman or Snake
    It's simply define by subclassing the trigger class, ie a non-stationary non-blocking object with a callback on collision
    """



    # def step(self):
    #     """Step function triggered by the environment
    #        - We look if the an agent has encountered the collectible
    #        - The on_collision method is called that need to be subclassed
    #        - The object is removed from the environment
    #     """

    #     # Find collisions
    #     is_collision,objects = self.collides_with(self.env.objects)
    #     assert len(objects) <= 1
        
    #     # Trigger callback
    #     if is_collision:
    #         self.on_collision(objects)

    #         # Remove object from the environment
    #         self.env.remove_object(self)



    # def on_collision(self,obj):
    #     print(f"Collision with objects {objects}, this function should be overridden")


    def on_trigger(self,obj):
        pass

    def on_trigger_exit(self,obj):
        pass


    def on_collision(self,obj):

        self.on_trigger(obj)
        self.env.remove_object(self)
        self.on_trigger_exit(obj)